/*
 * MPTCP implementation - OPPORTUNISTIC LINKED INCREASES CONGESTION CONTROL:
 *
 * Algorithm design: 
 * Ramin Khalili <ramin.khalili@epfl.ch> 
 * Nicolas Gast <nicolas.gast@epfl.ch> 
 * Jean-Yves Le Boudec <jean-yves.leboudec@epfl.ch>
 *
 * Implementation: 
 * Ramin Khalili <ramin.khalili@epfl.ch> 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */


#include <net/tcp.h>
#include <net/mptcp.h>

#include <linux/module.h>

static int scale = 10; 
static inline u64 mptcp_ccc_scale(u32 val, int scale)
{
	return (u64) val << scale;
}

/* take care of artificially inflate (see RFC5681) 
    of cwnd during fast-retransmit phase*/ 
u32 mptcp_get_crt_cwnd(struct tcp_sock *tp)
{
	struct inet_connection_sock *icsk = inet_csk((struct sock *) tp);

	if (icsk->icsk_ca_state == TCP_CA_Recovery)
		return tp->snd_ssthresh;
	else
		return tp->snd_cwnd;
}

/* return the dominator of the first term of 
   the increasing term */
u64 mptcp_get_rate(struct mptcp_cb* mpcb , u64 path_rtt) 
{
	struct tcp_sock *tp;
	struct sock *sub_sk;
	u64 rate = 0, scaled_num = 0;
	u32 tmp_cwnd = 1;
	mptcp_for_each_sk(mpcb, sub_sk) {
		u64 rtt=1;
		if (!mptcp_sk_can_send(sub_sk))
			continue;
		if (tp->srtt > 0)
			rtt = tp->srtt;
		tmp_cwnd = mptcp_get_crt_cwnd(tp);
		scaled_num = mptcp_ccc_scale(tmp_cwnd, scale) * path_rtt;
		rate += div_u64(scaled_num , rtt);
	}
	rate *= rate;
	return rate;
}

/* find the maximum cwnd, used to find set M*/
u32 mptcp_get_max_cwnd(struct mptcp_cb* mpcb) 
{
	struct tcp_sock *tp;
	struct sock *sub_sk;
	u32 best_cwnd = 1, tmp_cwnd = 1;

	mptcp_for_each_sk(mpcb,sub_sk) {
		if (!mptcp_sk_can_send(sub_sk))
			continue;
		tmp_cwnd = mptcp_get_crt_cwnd(tp);		
		if(tmp_cwnd >= best_cwnd)
			best_cwnd = tmp_cwnd;
	}
	return best_cwnd;
}


/* will retun back the value of epsilon */
void mptcp_get_epsilon(struct mptcp_cb* mpcb) 
{
	struct tcp_sock *tp;
	struct sock *sub_sk;
	u64 tmp_int=0, tmp_rtt=1, best_int=0, best_rtt=1;
	u32 max_cwnd = 1, best_cwnd = 1, tmp_cwnd = 1;
	u8 M=0, B_not_M=0;

	max_cwnd = mptcp_get_max_cwnd(mpcb);

	/* find the best path */
	mptcp_for_each_sk(mpcb, sub_sk) {
		tmp_rtt = 1;
		if (!mptcp_sk_can_send(sub_sk))
			continue;

		if (tp->srtt > 0)
			tmp_rtt = tp->srtt * tp->srtt;		
		tmp_int = max(tp->mptcp_loss3-tp->mptcp_loss2,
			      tp->mptcp_loss2-tp->mptcp_loss1);
		tmp_cwnd = mptcp_get_crt_cwnd(tp);		
		if(tmp_int * best_rtt >= best_int * tmp_rtt){
			best_rtt = tmp_rtt;
			best_int = tmp_int;
			best_cwnd = tmp_cwnd;
		}
	}

	/* find the size of M and B_not_M */
	mptcp_for_each_sk(mpcb, sub_sk) {
		tmp_rtt = 1;
		if (!mptcp_sk_can_send(sub_sk))
			continue;
		tmp_cwnd = mptcp_get_crt_cwnd(tp);				
		if(tmp_cwnd == max_cwnd){
			M++;
		}else{ 
			if (tp->srtt > 0)
				tmp_rtt = tp->srtt * tp->srtt;
			tmp_int = max(tp->mptcp_loss3-tp->mptcp_loss2,
				      tp->mptcp_loss2-tp->mptcp_loss1);
			if(tmp_int * best_rtt == best_int * tmp_rtt)
				B_not_M++;
		}
	}

	/* check if the path is in M or B_not_M and 
	   set the value of epsilon accordingly */
	mptcp_for_each_sk(mpcb, sub_sk) {
		tmp_rtt = 1;
		if (!mptcp_sk_can_send(sub_sk))
			continue;
		
		if(B_not_M == 0){
			tp->epsilon_num = 0;
			tp->epsilon_den = 1;
		}else{
			if (tp->srtt > 0)
				tmp_rtt = tp->srtt * tp->srtt;
			tmp_int = max(tp->mptcp_loss3-tp->mptcp_loss2,
				      tp->mptcp_loss2-tp->mptcp_loss1);
		        tmp_cwnd = mptcp_get_crt_cwnd(tp);		
			if(tmp_cwnd < max_cwnd && 
			   tmp_int * best_rtt == best_int * tmp_rtt){
				tp->epsilon_num = 1;
				tp->epsilon_den = mpcb->cnt_established * B_not_M;
			}else if(tmp_cwnd == max_cwnd){
				tp->epsilon_num = -1;
				tp->epsilon_den = mpcb->cnt_established * M;
			}else{
				tp->epsilon_num = 0;
				tp->epsilon_den = 1;
			}
		}				
	}

}

/* setting the initial values */
static void mptcp_cc_init(struct sock *sk)
{
	struct tcp_sock *tp = tcp_sk(sk);

	if (mpcb_from_tcpsock(tcp_sk(sk))) {
		tp->mptcp_loss1 = tp->snd_una;
		tp->mptcp_loss2 = tp->snd_una;
		tp->mptcp_loss3 = tp->snd_una;
		tp->mptcp_ssthresh = tp->snd_ssthresh;
		tp->mptcp_previous_cwnd = 1;
		tp->mptcp_snd_cwnd_cnt = 0;
		tp->epsilon_num = 0;
		tp->epsilon_den = 1;
	}
}

/* updating inter-loss distance and ssthresh */
static void mptcp_state(struct sock *sk, u8 new_state)
{

	if (!mpcb_from_tcpsock(tcp_sk(sk)))
		return;

	if (new_state == TCP_CA_Loss ||
	    new_state == TCP_CA_Recovery || new_state == TCP_CA_CWR){
		struct tcp_sock *tp = tcp_sk(sk);
	        struct mptcp_cb *mpcb = mpcb_from_tcpsock(tp);
		struct inet_connection_sock *icsk = inet_csk((struct sock *) tp);
		if ((tp->mptcp_loss3 != tp->mptcp_loss2) 
		     && !icsk->icsk_retransmits) 
		{	
			tp->mptcp_loss1=tp->mptcp_loss2;
			tp->mptcp_loss2=tp->mptcp_loss3;
			if (mpcb->cnt_established == 1)
				tp->mptcp_ssthresh = 
				    max(tp->mptcp_previous_cwnd >> 1U, 2U);
			else
				tp->mptcp_ssthresh = 
				    max(tp->mptcp_previous_cwnd >> 1U, 1U);
		}
	}

}

/* main algorithm */
static void mptcp_fc_cong_avoid(struct sock *sk, u32 ack, u32 in_flight)
{
	struct tcp_sock *tp = tcp_sk(sk), *meta_tp;
	struct mptcp_cb *mpcb = mpcb_from_tcpsock(tp);
	u64 inc_num = 0, inc_den = 0, rate = 0, rtt = 1, cwnd_scaled = 1;
	u32 cwnd = 1; 
	u32 ssthresh = min(tp->snd_ssthresh , tp->mptcp_ssthresh);
	if (mpcb->cnt_established > 1)
		ssthresh--;
	tp->mptcp_loss3 = tp->snd_una;

        mptcp_debug("%d %d %d %d %d\n", 
		     tp->snd_cwnd, ssthresh, 
		     tp->epsilon_num, tp->epsilon_den, tcp_time_stamp);

	if (!mpcb || !tp->mpc) {
		tcp_reno_cong_avoid(sk, ack, in_flight);	
		tp->mptcp_previous_cwnd=tp->snd_cwnd;		
		return;
	}

	meta_tp = mpcb_meta_tp(mpcb);
	if(!tcp_is_cwnd_limited(sk, in_flight)){
		tp->mptcp_previous_cwnd=tp->snd_cwnd;
		return;
	}

	/* slow start if it is in the safe area */
	if (tp->snd_cwnd <= ssthresh){	
		tcp_slow_start(tp);
		tp->mptcp_previous_cwnd=tp->snd_cwnd;
		return;
	}

	if (tp->srtt > 0)
		rtt = tp->srtt;
	mptcp_get_epsilon(mpcb);
	rate = mptcp_get_rate(mpcb,rtt);
	cwnd_scaled = mptcp_ccc_scale(tp->snd_cwnd, scale);
	inc_den = tp->epsilon_den * tp->snd_cwnd * rate;

	/* calculate the increasing term, scaling 
	   is used to reduce the rounding effect */
	if (tp->epsilon_num == -1){
		if (tp->epsilon_den * cwnd_scaled * cwnd_scaled < rate){
			inc_num = rate - tp->epsilon_den *
			    cwnd_scaled * cwnd_scaled;
			tp->mptcp_snd_cwnd_cnt -= div64_u64(
			    mptcp_ccc_scale(inc_num , scale) , inc_den);
		}else{
			inc_num = tp->epsilon_den * 
			    cwnd_scaled * cwnd_scaled - rate;		
			tp->mptcp_snd_cwnd_cnt += div64_u64(
			    mptcp_ccc_scale(inc_num , scale) , inc_den);
		}
	}else{
		inc_num = tp->epsilon_num * rate + 
	            tp->epsilon_den * cwnd_scaled * cwnd_scaled;		
		tp->mptcp_snd_cwnd_cnt += div64_u64(
		    mptcp_ccc_scale(inc_num , scale) , inc_den);
	}


	if (sysctl_tcp_abc) {
		if (tp->bytes_acked >= tp->snd_cwnd * meta_tp->mss_cache) {
			tp->bytes_acked -= tp->snd_cwnd * meta_tp->mss_cache;
			if (tp->snd_cwnd < tp->snd_cwnd_clamp)
				tp->snd_cwnd++;
		}
	} else {
		if (tp->mptcp_snd_cwnd_cnt >= 1023){
			if (tp->snd_cwnd < tp->snd_cwnd_clamp)
				tp->snd_cwnd++;
			tp->mptcp_snd_cwnd_cnt = 0;
		}else if (tp->mptcp_snd_cwnd_cnt <= -1023){
			if (tp->snd_cwnd < tp->snd_cwnd_clamp)
				tp->snd_cwnd = max(1 , tp->snd_cwnd-1);
			tp->mptcp_snd_cwnd_cnt = 0;
		}
	}
	tp->mptcp_previous_cwnd=tp->snd_cwnd;

}

static struct tcp_congestion_ops mptcp_fc = {
	.init		= mptcp_cc_init,
	.ssthresh	= tcp_reno_ssthresh,
	.cong_avoid	= mptcp_fc_cong_avoid,
	.set_state	= mptcp_state,
	.min_cwnd	= tcp_reno_min_cwnd,
	.owner		= THIS_MODULE,
	.name		= "coupled",
};

static int __init mptcp_fc_register(void)
{
	return tcp_register_congestion_control(&mptcp_fc);
}

static void __exit mptcp_fc_unregister(void)
{
	tcp_unregister_congestion_control(&mptcp_fc);
}

module_init(mptcp_fc_register);
module_exit(mptcp_fc_unregister);

MODULE_AUTHOR("Ramin Khalili, Nicolas Gast, Jean-Yves Le Boudec");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MPTCP COUPLED CONGESTION CONTROL");
MODULE_VERSION("0.1");
